#!/usr/bin/env python
import hashlib
import argparse
import re
import os
import subprocess
import errno

#tool to take list of URI's, and create symlinks in SNAKEMAKE_SINGULARITY_DIR



def hash(uri):
        md5hash = hashlib.md5()
        md5hash.update(uri.encode())
        return md5hash.hexdigest()

def get_cached_path(uri):

#parse uri with regular exp to get org, repo, tag:
# group 1 is docker or shub
# group 2 is org (any non-whitespace string)
# group 3 is repo
# group 4 is tag
 
    m = re.search('(?P<hub>\w+)://(?P<org>\S+)/(?P<repo>\S+):(?P<tag>\S+)',uri)
    existing_cache = '{org}_{repo}_{tag}.sif'.format(**m.groupdict())

    singularity_dir = os.environ['SINGULARITY_DIR']
    if m.group('hub') == 'docker':
        pulled_path = os.path.join(singularity_dir,'bids-apps','{org}_{repo}_{tag}.sif'.format(**m.groupdict()))
    elif m.group('hub') == 'shub':
        pulled_path = os.path.join(singularity_dir,'{org}-{repo}-master-{tag}.sif'.format(**m.groupdict()))
    else:
        print('must use docker:// or shub://')

    return pulled_path

def get_linked_path(uri,snakemake_singularity_dir):
    linked_path = os.path.join(snakemake_singularity_dir,hash(uri)+'.simg')
    return linked_path

def make_sym_link(uri,snakemake_singularity_dir):
    cached = get_cached_path(uri)
    linked = get_linked_path(uri,snakemake_singularity_dir )
    if not os.path.exists(cached):
        print(f'ERROR: image {uri} has not already been pulled to {cached}')
    else:
        if os.path.exists(linked):
            print(f'WARNING: link to {cached} already exists in {linked}')
        else:
            subprocess.run(f'ln -s {cached} {linked}',shell=True)
            print(f'linking {cached} in {linked}')

    
#snakemake_singularity_dir = os.environ['SNAKEMAKE_SINGULARITY_DIR']
if 'SNAKEMAKE_SINGULARITY_DIR' in os.environ:
    default_snakemake_singularity_dir = os.environ['SNAKEMAKE_SINGULARITY_DIR']
else:
    default_snakemake_singularity_dir = os.path.join(os.path.expanduser('~'),'.snakemake_singularity')


parser = argparse.ArgumentParser()
group = parser.add_mutually_exclusive_group(required=True)
group.add_argument('--uri', 
                    help='uri of container, {shub/docker}://org/repo:tag')
group.add_argument('--txtfile',#type=argparse.FileType('r'), 
                    help='text file containing uri on each line')
parser.add_argument('--snakemake-singularity-dir',
                    help='Folder to create symlinks to cached singularity images. '
                            'Pass this to snakemake with the --singularity-prefix option,'
                            ' or use a profile that sets this option (e.g. cc-slurm).  '
                            'If set, the SNAKEMAKE_SINGULARITY_DIR environment variable '
                            'will be used for this.  '
                            'Default: {}'.format(default_snakemake_singularity_dir),
                            default=default_snakemake_singularity_dir)

args = parser.parse_args()

#create snakemake_singularity_dir

try:
    os.mkdir(args.snakemake_singularity_dir)
except OSError as exc:
    if exc.errno != errno.EEXIST:
        raise
    pass




if args.uri is not None:
    make_sym_link(args.uri,args.snakemake_singularity_dir)
else:
    #loop through lines of txtfile
    print(args.txtfile)
    with open(args.txtfile) as f:
        for line in f:
            make_sym_link(line.rstrip(), args.snakemake_singularity_dir)

    
   




from __future__ import absolute_import

from typing import Any, Callable, Generic, Optional, TypeVar

import attrs
from typing_extensions import Self

from kslurm.args.arg import Context, Parser

T = TypeVar("T")


class NO_FORMAT:
    def __new__(cls, arg: Any, __: Any, ___: Any):
        return arg


class NO_CONVERT:
    def __new__(cls, arg: Any):
        return arg


@attrs.define
class Action(Generic[T]):
    formatter: Callable[[str, Parser[Any], Context], T] = NO_FORMAT
    validator: Optional[Callable[[str, Parser[Any], Context], None]] = None

    @property
    def has_formatter(self):
        return self.formatter is not NO_FORMAT and self.formatter is not NO_CONVERT

    def with_converter(self, _converter: Callable[[str], Any], /) -> Self:
        return self.convert(_converter)

    def convert(self, converter: Callable[[str], T]) -> Self:
        return convert(converter, self)

    def format(self, formatter: Callable[[str, Parser[T], Context], T]) -> Self:
        return format(formatter, self)

    def validate(self, validator: Callable[[str], None]) -> Self:
        return validate(validator, self)

    def replace(self):
        return replace(formatter=self.formatter, validator=self.validator)

    def append(self):
        return append(formatter=self.formatter, validator=self.validator)

    def raises(self, error: Exception):
        return raises(error=error, validator=self.validator)


class replace(Action[T]):
    def __call__(self, arg: str, parser: Parser[Any], context: Context):
        if self.validator is not None:
            self.validator(arg, parser, context)
        return self.formatter(arg, parser, context)


class append(Action[T]):
    def __call__(self, arg: str, parser: Parser[list[Any]], context: Context):
        if self.validator is not None:
            self.validator(arg, parser, context)
        curr_value = parser.value or []
        return curr_value + [self.formatter(arg, parser, context)]


@attrs.define
class raises:
    error: Exception
    validator: Optional[Callable[[str, Parser[Any], Context], None]] = None

    @property
    def has_formatter(self):
        return True

    def with_converter(self, _converter: Callable[[str], Any], /) -> Self:
        raise NotImplementedError()

    def __call__(self, arg: str, parser: Parser[Any], context: Context):
        if self.validator is not None:
            self.validator(arg, parser, context)
        raise self.error


def convert(converter: Callable[[str], T], template: Optional[Action[T]] = None):
    def formatter(arg: str, _: Any, __: Any):
        return converter(arg)

    newformatter = NO_FORMAT if converter is NO_CONVERT else formatter

    return format(newformatter, template)


def format(
    formatter: Callable[[str, Parser[T], Context], T],
    template: Optional[Action[T]] = None,
):
    if template is not None:
        template.formatter = formatter
    else:
        template = Action(formatter=formatter, validator=None)
    return template


def validate(
    validator: Callable[[str], None],
    template: Optional[Action[T]] = None,
) -> Action[T]:
    def _validator(arg: str, _: Any, __: Any):
        validator(arg)

    if template is not None:
        template.validator = _validator
    else:
        template = Action(validator=_validator)
    return template



from __future__ import absolute_import, annotations

import hashlib
from typing import TypeVar, Union

T = TypeVar("T")


def get_hash(item: Union[str, bytes], method: str = "md5"):
    if isinstance(item, str):
        item = item.encode()
    if method == "md5":
        return hashlib.md5(item).hexdigest()
    elif method == "sha512":
        return hashlib.sha512(item).hexdigest()
    else:
        raise TypeError(f"method '{method}' is not a valid hash method")



from __future__ import absolute_import, annotations

import json
from collections import UserDict
from pathlib import Path

import appdirs

from kslurm.args.command import CommandError

CONFIG_PATH = Path(appdirs.user_config_dir("kslurm"), "config.json")


class Config(UserDict[str, str]):
    def __init__(self):
        self._path = CONFIG_PATH
        if not self._path.exists():
            self._path.parent.mkdir(parents=True, exist_ok=True)
            with self._path.open("w") as f:
                json.dump({}, f)
                self.data = {}
        else:
            with self._path.open("r") as f:
                self.data = json.load(f)

    def write(self):
        with self._path.open("w") as f:
            json.dump(self.data, f)

    def get_children(self, entity: str):
        for key, value in self.data.items():
            if key.startswith(entity + "."):
                yield key, value

    def __str__(self):
        return "• " + "\n• ".join(self.data.keys()) if self.data else ""


class InvalidPipdirError(CommandError):
    pass


class PipDir(type(Path())):
    def __new__(cls):
        pipdir = Config().get("pipdir")
        if pipdir is None:
            raise InvalidPipdirError(
                "pipdir not set. Please set pipdir using `kslurm config pipdir "
                "<directory>`, typically to a project-space or permanent storage "
                "directory"
            )
        pipdir = Path(pipdir)
        try:
            pipdir.mkdir(exist_ok=True)
        except FileNotFoundError:
            raise InvalidPipdirError(
                f"Unable to create pipdir in non-existant directory '{pipdir.parent}'. "
                "Please manually create this directory."
            )
        return super().__new__(cls, pipdir)
